@@ -49,8 +49,13 @@ class UserEventFactory(object):
         if activated_experiment:
             experiment_id = activated_experiment.id
 
-        if variation_id and experiment_id:
+        if variation_id and flag_key:
+            # need this condition when we send events involving forced decisions
+            # (F-to-D or E-to-D with any ruleKey/variationKey combinations)
+            variation = project_config.get_flag_variation(flag_key, 'id', variation_id)
+        elif variation_id and experiment_id:
             variation = project_config.get_variation_from_id_by_experiment_id(experiment_id, variation_id)
+
         event_context = user_event.EventContext(
             project_config.account_id, project_config.project_id, project_config.revision, project_config.anonymize_ip,
         )
